class KERR:
    WEBSOCKET_ERR=100
    SERIAL_COMM_ERROR=101
    CANNOT_FIND_DISK=102
    UPLOAD_IMAGE_FAIL=103


    UPLOAD_FIRM_FAIL=110
    CANNOT_FIND_COMMPORT=170
    CANNOT_FIND_COMMPORT_BUTDISK=171
    DISK_IN_BOOTLOADER_MODE=172
    GO_NORMAL_REPL_FAIL=175
    EXEC_COMMAND_FAIL=176
    DRIVER_INSTALL_FAIL = 180
    DRIVER_INSTALL_DOWNLOADFAIL = 181
    DRIVER_INSTALL_OSX = 183
    LIST_FILE_FAIL=150
    GET_DISK_INFO_FAIL=160

    SYNC_FILE_FAIL=165


    CANNT_USE_GETOBJECT = 189
    UNKNOW_ERROR = 188


